import json
import logging

from datetime import datetime
from typing import Any

import dbus_fast.aio as dbus_aio
import dbus_fast.introspection as dbus_introspection

from dbus2mqtt import AppContext
from dbus2mqtt.config import InterfaceConfig, SubscriptionConfig
from dbus2mqtt.dbus.dbus_types import BusNameSubscriptions, SubscribedInterface
from dbus2mqtt.dbus.dbus_util import (
    camel_to_snake,
    unwrap_dbus_object,
    unwrap_dbus_objects,
)
from dbus2mqtt.dbus.introspection_patches.mpris_playerctl import (
    mpris_introspection_playerctl,
)
from dbus2mqtt.dbus.introspection_patches.mpris_vlc import mpris_introspection_vlc
from dbus2mqtt.event_broker import DbusSignalWithState, MqttMessage
from dbus2mqtt.flow.flow_processor import FlowScheduler, FlowTriggerMessage

logger = logging.getLogger(__name__)


class DbusClient:

    def __init__(self, app_context: AppContext, bus: dbus_aio.message_bus.MessageBus, flow_scheduler: FlowScheduler):
        self.app_context = app_context
        self.config = app_context.config.dbus
        self.event_broker = app_context.event_broker
        self.templating = app_context.templating
        self.bus = bus
        self.flow_scheduler = flow_scheduler
        self.subscriptions: dict[str, BusNameSubscriptions] = {}

    async def connect(self):

        if not self.bus.connected:
            await self.bus.connect()

            if self.bus.connected:
                logger.info(f"Connected to {self.bus._bus_address}")
            else:
                logger.warning(f"Failed to connect to {self.bus._bus_address}")

            introspection = await self.bus.introspect('org.freedesktop.DBus', '/org/freedesktop/DBus')
            obj = self.bus.get_proxy_object('org.freedesktop.DBus', '/org/freedesktop/DBus', introspection)
            dbus_interface = obj.get_interface('org.freedesktop.DBus')

            # subscribe to NameOwnerChanged which allows us to detect new bus_names
            dbus_interface.__getattribute__("on_name_owner_changed")(self._dbus_name_owner_changed_callback)

            # subscribe to existing registered bus_names we are interested in
            connected_bus_names = await dbus_interface.__getattribute__("call_list_names")()

            new_subscriped_interfaces: list[SubscribedInterface] = []
            for bus_name in connected_bus_names:
                new_subscriped_interfaces.extend(await self._handle_bus_name_added(bus_name))

            logger.info(f"subscriptions on startup: {list(set([si.bus_name for si in new_subscriped_interfaces]))}")

    def get_proxy_object(self, bus_name: str, path: str) -> dbus_aio.proxy_object.ProxyObject | None:

        bus_name_subscriptions = self.subscriptions.get(bus_name)
        if bus_name_subscriptions:
            proxy_object = bus_name_subscriptions.path_objects.get(path)
            if proxy_object:
                return proxy_object

        return None

    def _ensure_proxy_object_subscription(self, bus_name: str, path: str, introspection: dbus_introspection.Node):

        bus_name_subscriptions = self.subscriptions.get(bus_name)
        if not bus_name_subscriptions:
            bus_name_subscriptions = BusNameSubscriptions(bus_name)
            self.subscriptions[bus_name] = bus_name_subscriptions

        proxy_object = bus_name_subscriptions.path_objects.get(path)
        if not proxy_object:
            proxy_object = self.bus.get_proxy_object(bus_name, path, introspection)
            bus_name_subscriptions.path_objects[path] = proxy_object

        return proxy_object, bus_name_subscriptions

    def _dbus_fast_signal_publisher(self, dbus_signal_state: dict[str, Any], *args):
        unwrapped_args = unwrap_dbus_objects(args)
        self.event_broker.on_dbus_signal(
            DbusSignalWithState(**dbus_signal_state, args=unwrapped_args)
        )

    def _dbus_fast_signal_handler(self, signal: dbus_introspection.Signal, state: dict[str, Any]) -> Any:
        expected_args = len(signal.args)

        if expected_args == 1:
            return lambda a: self._dbus_fast_signal_publisher(state, a)
        elif expected_args == 2:
            return lambda a, b: self._dbus_fast_signal_publisher(state, a, b)
        elif expected_args == 3:
            return lambda a, b, c: self._dbus_fast_signal_publisher(state, a, b, c)
        elif expected_args == 4:
            return lambda a, b, c, d: self._dbus_fast_signal_publisher(state, a, b, c, d)
        raise ValueError("Unsupported nr of arguments")

    async def _subscribe_interface(self, bus_name: str, path: str, introspection: dbus_introspection.Node, interface: dbus_introspection.Interface, subscription_config: SubscriptionConfig, si: InterfaceConfig) -> SubscribedInterface:

        proxy_object, bus_name_subscriptions = self._ensure_proxy_object_subscription(bus_name, path, introspection)
        obj_interface = proxy_object.get_interface(interface.name)

        interface_signals = dict((s.name, s) for s in interface.signals)

        logger.debug(f"subscribe: bus_name={bus_name}, path={path}, interface={interface.name}, proxy_interface: signals={list(interface_signals.keys())}")

        for signal_config in si.signals:
            interface_signal = interface_signals.get(signal_config.signal)
            if interface_signal:

                on_signal_method_name = "on_" + camel_to_snake(signal_config.signal)
                dbus_signal_state = {
                    "bus_name": bus_name,
                    "path": path,
                    "interface_name": interface.name,
                    "subscription_config": subscription_config,
                    "signal_config": signal_config,
                }

                handler = self._dbus_fast_signal_handler(interface_signal, dbus_signal_state)
                obj_interface.__getattribute__(on_signal_method_name)(handler)
                logger.info(f"subscribed with signal_handler: signal={signal_config.signal}, bus_name={bus_name}, path={path}, interface={interface.name}")

            else:
                logger.warning(f"Invalid signal: signal={signal_config.signal}, bus_name={bus_name}, path={path}, interface={interface.name}")

        return SubscribedInterface(
            bus_name=bus_name,
            path=path,
            interface_name=interface.name,
            subscription_config=subscription_config,
            interface_config=si
        )

    async def _process_interface(self, bus_name: str, path: str, introspection: dbus_introspection.Node, interface: dbus_introspection.Interface) -> list[SubscribedInterface]:

        logger.debug(f"process_interface: {bus_name}, {path}, {interface.name}")
        subscription_configs = self.config.get_subscription_configs(bus_name, path)
        new_subscriptions: list[SubscribedInterface] = []
        for subscription in subscription_configs:
            logger.debug(f"processing subscription config: {subscription.bus_name}, {subscription.path}")
            for subscription_interface in subscription.interfaces:
                if subscription_interface.interface == interface.name:
                    logger.debug(f"matching config found for bus_name={bus_name}, path={path}, interface={interface.name}")
                    subscribed_iterface = await self._subscribe_interface(bus_name, path, introspection, interface, subscription, subscription_interface)

                    new_subscriptions.append(subscribed_iterface)

        return new_subscriptions

    async def _introspect(self, bus_name: str, path: str) -> dbus_introspection.Node:

        if path == "/org/mpris/MediaPlayer2" and bus_name.startswith("org.mpris.MediaPlayer2.vlc"):
            # vlc 3.x branch contains an incomplete dbus introspection
            # https://github.com/videolan/vlc/commit/48e593f164d2bf09b0ca096d88c86d78ec1a2ca0
            # Until vlc 4.x is out we use the official specification instead
            introspection = mpris_introspection_vlc
        else:
            introspection = await self.bus.introspect(bus_name, path)

        # MPRIS: If no introspection data is available, load a default
        if path == "/org/mpris/MediaPlayer2" and bus_name.startswith("org.mpris.MediaPlayer2.") and len(introspection.interfaces) == 0:
            introspection = mpris_introspection_playerctl

        return introspection

    async def _visit_bus_name_path(self, bus_name: str, path: str) -> list[SubscribedInterface]:

        new_subscriptions: list[SubscribedInterface] = []

        try:
            introspection = await self._introspect(bus_name, path)
        except TypeError as e:
            logger.warning(f"bus.introspect failed, bus_name={bus_name}, path={path}: {e}")
            return new_subscriptions

        if len(introspection.nodes) == 0:
            logger.debug(f"leaf node: bus_name={bus_name}, path={path}, is_root={introspection.is_root}, interfaces={[i.name for i in introspection.interfaces]}")

        for interface in introspection.interfaces:
            new_subscriptions.extend(
                await self._process_interface(bus_name, path, introspection, interface)
            )

        for node in introspection.nodes:
            path_seperator = "" if path.endswith('/') else "/"
            new_subscriptions.extend(
                await self._visit_bus_name_path(bus_name, f"{path}{path_seperator}{node.name}")
            )

        return new_subscriptions

    async def _handle_bus_name_added(self, bus_name: str) -> list[SubscribedInterface]:

        if not self.config.is_bus_name_configured(bus_name):
            return []

        # sanity checks
        for umh in self.bus._user_message_handlers:
            umh_bus_name = umh.__self__.bus_name
            if umh_bus_name == bus_name:
                logger.warning(f"handle_bus_name_added: {umh_bus_name} already added")

        new_subscriped_interfaces = await self._visit_bus_name_path(bus_name, "/")

        logger.info(f"new_subscriptions: {list(set([si.bus_name for si in new_subscriped_interfaces]))}")

        # setup and process triggers for each flow in each subscription
        processed_new_subscriptions: set[str] = set()

        # With all subscriptions in place, we can now ensure schedulers are created
        # create a FlowProcessor per bus_name/path subscription?
        # One global or a per subscription FlowProcessor.flow_processor_task?
        # Start a new timer job, but leverage existing FlowScheduler
        # How does the FlowScheduler now it should invoke the local FlowPocessor?
        # Maybe use queues to communicate from here with the FlowProcessor?
        # e.g.: StartFlows, StopFlows,

        for subscribed_interface in new_subscriped_interfaces:

            subscription_config = subscribed_interface.subscription_config
            if subscription_config.id not in processed_new_subscriptions:

                # Ensure all schedulers are started
                self.flow_scheduler.start_flow_set(subscription_config.flows)

                # Trigger flows that have a bus_name_added trigger configured
                await self._trigger_bus_name_added(subscription_config, subscribed_interface.bus_name, subscribed_interface.path)

                processed_new_subscriptions.add(subscription_config.id)

        return new_subscriped_interfaces

    async def _trigger_bus_name_added(self, subscription_config: SubscriptionConfig, bus_name: str, path: str):

        for flow in subscription_config.flows:
            for trigger in flow.triggers:
                if trigger.type == "bus_name_added":
                    trigger_context = {
                        "bus_name": bus_name,
                        "path": path
                    }
                    trigger_message = FlowTriggerMessage(
                        flow,
                        trigger,
                        datetime.now(),
                        trigger_context=trigger_context
                    )
                    await self.event_broker.flow_trigger_queue.async_q.put(trigger_message)

    async def _handle_bus_name_removed(self, bus_name: str):

        bus_name_subscriptions = self.subscriptions.get(bus_name)

        if bus_name_subscriptions:
            for path, proxy_object in bus_name_subscriptions.path_objects.items():

                subscription_configs = self.config.get_subscription_configs(bus_name=bus_name, path=path)
                for subscription_config in subscription_configs:

                    # Trigger flows that have a bus_name_added trigger configured
                    await self._trigger_bus_name_removed(subscription_config, bus_name, path)

                    # Stop schedule triggers
                    self.flow_scheduler.stop_flow_set(subscription_config.flows)

                # Wait for completion
                await self.event_broker.flow_trigger_queue.async_q.join()

                # clean up all dbus matchrules
                for interface in proxy_object._interfaces.values():
                    proxy_interface: dbus_aio.proxy_object.ProxyInterface = interface

                    # officially you should do 'off_...' but the below is easier
                    # proxy_interface.off_properties_changed(self.on_properties_changed)

                    # clean lingering interface matchrule from bus
                    if proxy_interface._signal_match_rule in self.bus._match_rules.keys():
                        self.bus._remove_match_rule(proxy_interface._signal_match_rule)

                    # clean lingering interface messgage handler from bus
                    self.bus.remove_message_handler(proxy_interface._message_handler)

            del self.subscriptions[bus_name]

    async def _trigger_bus_name_removed(self, subscription_config: SubscriptionConfig, bus_name: str, path: str):

        # Trigger flows that have a bus_name_removed trigger configured
        for flow in subscription_config.flows:
            for trigger in flow.triggers:
                if trigger.type == "bus_name_removed":
                    trigger_context = {
                        "bus_name": bus_name,
                        "path": path
                    }
                    trigger_message = FlowTriggerMessage(
                        flow,
                        trigger,
                        datetime.now(),
                        trigger_context=trigger_context
                    )
                    await self.event_broker.flow_trigger_queue.async_q.put(trigger_message)

    async def _dbus_name_owner_changed_callback(self, name, old_owner, new_owner):

        logger.debug(f'NameOwnerChanged: name=q{name}, old_owner={old_owner}, new_owner={new_owner}')

        if new_owner and not old_owner:
            logger.debug(f'NameOwnerChanged.new: name={name}')
            await self._handle_bus_name_added(name)
        if old_owner and not new_owner:
            logger.debug(f'NameOwnerChanged.old: name={name}')
            await self._handle_bus_name_removed(name)

    async def call_dbus_interface_method(self, interface: dbus_aio.proxy_object.ProxyInterface, method: str, method_args: list[Any]):

        call_method_name = "call_" + camel_to_snake(method)
        res = await interface.__getattribute__(call_method_name)(*method_args)

        if res:
            res = unwrap_dbus_object(res)

        logger.debug(f"call_dbus_interface_method: bus_name={interface.bus_name}, interface={interface.introspection.name}, method={method}, res={res}")

        return res

    async def get_dbus_interface_property(self, interface: dbus_aio.proxy_object.ProxyInterface, property: str) -> Any:

        call_method_name = "get_" + camel_to_snake(property)
        res = await interface.__getattribute__(call_method_name)()

        if res:
            res = unwrap_dbus_object(res)

        logger.debug(f"get_dbus_interface_property: bus_name={interface.bus_name}, interface={interface.introspection.name}, property={property}, res={res}")

        return res

    async def set_dbus_interface_property(self, interface: dbus_aio.proxy_object.ProxyInterface, property: str, value: Any) -> None:

        call_method_name = "set_" + camel_to_snake(property)
        await interface.__getattribute__(call_method_name)(value)

        logger.info(f"set_dbus_interface_property: bus_name={interface.bus_name}, interface={interface.introspection.name}, property={property}, value={value}")

    async def mqtt_receive_queue_processor_task(self):
        """Continuously processes messages from the async queue."""
        while True:
            msg = await self.event_broker.mqtt_receive_queue.async_q.get()  # Wait for a message
            try:
                await self._on_mqtt_msg(msg)
            except Exception as e:
                logger.warning(f"mqtt_receive_queue_processor_task: Exception {e}", exc_info=True)
            finally:
                self.event_broker.mqtt_receive_queue.async_q.task_done()

    async def dbus_signal_queue_processor_task(self):
        """Continuously processes messages from the async queue."""
        while True:
            signal = await self.event_broker.dbus_signal_queue.async_q.get()  # Wait for a message
            await self._handle_on_dbus_signal(signal)
            self.event_broker.dbus_signal_queue.async_q.task_done()

    async def _handle_on_dbus_signal(self, signal: DbusSignalWithState):

        logger.debug(f"dbus_signal: signal={signal.signal_config.signal}, args={signal.args}, bus_name={signal.bus_name}, path={signal.path}, interface={signal.interface_name}")

        for flow in signal.subscription_config.flows:
            for trigger in flow.triggers:
                if trigger.type == "dbus_signal" and signal.signal_config.signal == trigger.signal:

                    try:

                        matches_filter = True
                        if signal.signal_config.filter is not None:
                            matches_filter = signal.signal_config.matches_filter(self.app_context.templating, *signal.args)

                        if matches_filter:
                            trigger_context = {
                                "bus_name": signal.bus_name,
                                "path": signal.path,
                                "interface": signal.interface_name,
                                "args": signal.args
                            }
                            trigger_message = FlowTriggerMessage(
                                flow,
                                trigger,
                                datetime.now(),
                                trigger_context=trigger_context
                            )

                            await self.event_broker.flow_trigger_queue.async_q.put(trigger_message)
                    except Exception as e:
                        logger.warning(f"dbus_signal_queue_processor_task: Exception {e}", exc_info=True)

    async def _on_mqtt_msg(self, msg: MqttMessage):
        # self.queue.put({
        #     "topic": topic,
        #     "payload": payload
        # })

        found_matching_topic = False
        for subscription_configs in self.config.subscriptions:
            for interface_config in subscription_configs.interfaces:
                # TODO, performance improvement
                mqtt_topic = interface_config.render_mqtt_command_topic(self.templating, {})
                found_matching_topic |= mqtt_topic == msg.topic

        if not found_matching_topic:
            return

        logger.debug(f"on_mqtt_msg: topic={msg.topic}, payload={json.dumps(msg.payload)}")
        matched_method = False
        matched_property = False

        payload_method = msg.payload.get("method")
        payload_method_args = msg.payload.get("args") or []

        payload_property = msg.payload.get("property")
        payload_value = msg.payload.get("value")

        if payload_method is None and (payload_property is None or payload_value is None):
            logger.info(f"on_mqtt_msg: Unsupported payload, missing 'method' or 'property/value', got method={payload_method}, property={payload_property}, value={payload_value} from {msg.payload}")
            return

        for [bus_name, bus_name_subscription] in self.subscriptions.items():
            for [path, proxy_object] in bus_name_subscription.path_objects.items():
                for subscription_configs in self.config.get_subscription_configs(bus_name=bus_name, path=path):
                    for interface_config in subscription_configs.interfaces:

                        for method in interface_config.methods:

                            # filter configured method, configured topic, ...
                            if method.method == payload_method:
                                interface = proxy_object.get_interface(name=interface_config.interface)
                                matched_method = True

                                try:
                                    logger.info(f"on_mqtt_msg: method={method.method}, args={payload_method_args}, bus_name={bus_name}, path={path}, interface={interface_config.interface}")
                                    await self.call_dbus_interface_method(interface, method.method, payload_method_args)
                                except Exception as e:
                                    logger.warning(f"on_mqtt_msg: method={method.method}, args={payload_method_args}, bus_name={bus_name} failed, exception={e}")

                        for property in interface_config.properties:
                            # filter configured property, configured topic, ...
                            if property.property == payload_property:
                                interface = proxy_object.get_interface(name=interface_config.interface)
                                matched_property = True

                                try:
                                    logger.info(f"on_mqtt_msg: property={property.property}, value={payload_value}, bus_name={bus_name}, path={path}, interface={interface_config.interface}")
                                    await self.set_dbus_interface_property(interface, property.property, payload_value)
                                except Exception as e:
                                    logger.warning(f"on_mqtt_msg: property={property.property}, value={payload_value}, bus_name={bus_name} failed, exception={e}")

        if not matched_method and not matched_property:
            if payload_method:
                logger.info(f"No configured or active dbus subscriptions for topic={msg.topic}, method={payload_method}, active bus_names={list(self.subscriptions.keys())}")
            if payload_property:
                logger.info(f"No configured or active dbus subscriptions for topic={msg.topic}, property={payload_property}, active bus_names={list(self.subscriptions.keys())}")

        # raw mode, payload contains: bus_name (specific or wildcard), path, interface_name
        # topic: dbus2mqtt/raw (with allowlist check)

        # predefined mode with topic matching from configuration
        # topic: dbus2mqtt/MediaPlayer/command




from unittest.mock import patch

import dbus_fast.aio as dbus_aio

from pydantic import SecretStr

from dbus2mqtt import AppContext, config
from dbus2mqtt.config import (
    FlowActionConfig,
    FlowConfig,
    FlowTriggerConfig,
    InterfaceConfig,
)
from dbus2mqtt.dbus.dbus_client import DbusClient
from dbus2mqtt.event_broker import EventBroker
from dbus2mqtt.flow.flow_processor import FlowProcessor, FlowScheduler
from dbus2mqtt.template.templating import TemplateEngine


def mocked_app_context():

    test_config = config.Config(dbus=config.DbusConfig(
        subscriptions=[
            config.SubscriptionConfig(
                bus_name="test.bus_name.*",
                path="/",
                interfaces=[
                    InterfaceConfig(
                        interface="test-interface-name"
                    )
                ]

            )
        ]
    ),
    mqtt=config.MqttConfig(
        host="localhost",
        username="test",
        password=SecretStr("test")
    ),
    flows=[]
    )

    event_broker = EventBroker()
    template_engine = TemplateEngine()
    app_context = AppContext(test_config, event_broker, template_engine)

    return app_context

def mocked_flow_processor(app_context: AppContext, trigger_config: FlowTriggerConfig, actions: list[FlowActionConfig]):

    flow_config = FlowConfig(triggers=[trigger_config], actions=actions)

    app_context.config.dbus.subscriptions[0].flows = [flow_config]

    processor = FlowProcessor(app_context)
    return processor, flow_config

def mocked_dbus_client(app_context: AppContext):

    with patch("socket.socket", autospec=True):

        bus = dbus_aio.message_bus.MessageBus(bus_address="unix:path=/test-path")
        flow_scheduler = FlowScheduler(app_context)

        dbus_client = DbusClient(app_context, bus, flow_scheduler)
        return dbus_client




import os

import dotenv

from dbus2mqtt.config import Config
from dbus2mqtt.config.jsonarparse import new_argument_parser

FILE_DIR = os.path.dirname(__file__)

def test_home_assistant_media_player_example():

    dotenv.load_dotenv(".env.example")

    parser = new_argument_parser()
    parser.add_class_arguments(Config)

    cfg = parser.parse_path(f"{FILE_DIR}/../../docs/examples/home_assistant_media_player.yaml")
    config = parser.instantiate_classes(cfg)

    assert config is not None

def test_linux_desktop_example():
    dotenv.load_dotenv(".env.example")

    parser = new_argument_parser()
    parser.add_class_arguments(Config)

    cfg = parser.parse_path(f"{FILE_DIR}/../../docs/examples/linux_desktop.yaml")
    config = parser.instantiate_classes(cfg)

    assert config is not None



from datetime import datetime

import pytest

from dbus2mqtt.config import FlowActionMqttPublishConfig, FlowTriggerScheduleConfig
from dbus2mqtt.flow.flow_processor import FlowTriggerMessage
from tests import mocked_app_context, mocked_flow_processor


@pytest.mark.asyncio
async def test_mqtt_publish_action():

    app_context = mocked_app_context()

    trigger_config = FlowTriggerScheduleConfig()
    processor, flow_config = mocked_flow_processor(app_context, trigger_config, actions=[
        FlowActionMqttPublishConfig(
            topic="dbus2mqtt/test",
            payload_template='{"test-key": "test-value"}'
        )
    ])

    await processor._process_flow_trigger(
        FlowTriggerMessage(flow_config, trigger_config, datetime.now())
    )

    mqtt_message = app_context.event_broker.mqtt_publish_queue.sync_q.get_nowait()

    assert mqtt_message is not None
    assert mqtt_message.payload_serialization_type == "json"

    payload: dict = mqtt_message.payload
    assert payload["test-key"] == "test-value"




from dbus2mqtt.template.templating import TemplateEngine


def test_preregisted_custom_function():

    # now is a custom dbus2mqtt function
    template = "now={{ now().isoformat() }}"

    templating = TemplateEngine()
    res = templating.render_template(template, object)

    assert isinstance(res, str)
    assert "now" in res

def test_non_template_str():
    templating = TemplateEngine()
    res = templating.render_template("str_value", object)
    assert res == "str_value"

def test_non_template_int_like():
    templating = TemplateEngine()
    res = templating.render_template("3", object)
    assert res == 3

def test_non_template_list_like():
    templating = TemplateEngine()
    res = templating.render_template("[1, 2, 3]", object)
    assert res == [1, 2, 3]

def test_str_template_str_result():
    templating = TemplateEngine()
    res = templating.render_template("{{ 'str_result' }}", object)
    assert res == "str_result"

def test_str_template_int_result():
    templating = TemplateEngine()
    res = templating.render_template("{{ 3 }}", object)
    assert res == 3

def test_str_template_list_result():
    templating = TemplateEngine()
    res = templating.render_template("{{ [1, 2, 3] }}", object)
    assert res == [1, 2, 3]

def test_str_template_int_result_as_str():
    templating = TemplateEngine()
    res = templating.render_template("{{ 3 }}", str)
    assert res == "3"

def test_none_result():
    templating = TemplateEngine()
    res = templating.render_template("{{ None }}", str)
    assert res is None

def test_dict_with_integer_expression():

    template = {
        "value": "{{ 1 | int }}",
    }

    templating = TemplateEngine()
    res = templating.render_template(template, dict)

    assert isinstance(res, dict)
    assert res["value"] == 1

def test_nested_dict_templates():

    nested_dict_template = {
        "a": 1,
        "b": {
            "c": "TestValueC"
        }
    }
    context = {
        "mpris_bus_names": ["org.mpris.MediaPlayer2.vlc", "org.mpris.MediaPlayer2.firefox"],
        "nested_dict_template": nested_dict_template
    }
    template = {
        "now": "{{ now().isoformat() }}",
        "dbus_names": "{{ mpris_bus_names }}",
        "nested_raw": nested_dict_template,
        "nested_template": "{{ nested_dict_template }}"
    }

    templating = TemplateEngine()
    res = templating.render_template(template, dict, context)

    assert isinstance(res, dict)

    test_dbus_names = res["dbus_names"]
    assert isinstance(test_dbus_names, list)
    assert len(test_dbus_names) == 2
    assert "org.mpris.MediaPlayer2.vlc" in res["dbus_names"]
    assert isinstance(res["now"], str)
    assert isinstance(res["nested_template"], dict)
    assert isinstance(res["nested_template"]["b"], dict)
    assert res["nested_template"]["b"]["c"] == "TestValueC"

def test_dict_template_with_quotes_and_newline():
    """
        Test that:
          1. a dict template with quotes and an ending newline is rendered correctly,
          2. dbus_call result with a nested dict is rendered correctly and convertible to a dict
    """
    custom_functions = {
        "dbus_list": lambda bus_name_pattern: ["org.mpris.MediaPlayer2.vlc", "org.mpris.MediaPlayer2.firefox"],
        "dbus_call": lambda bus_name, path, interface, method, method_args: {'Metadata': {}, 'Position': 0, 'CanControl': True},
        "dbus_property_get": lambda bus_name, path, interface, property, default_unsupported: {"value": 1}
    }

    templating = TemplateEngine()
    templating.add_functions(custom_functions)

    template = {
        "player_properties": "{{ dbus_call(mpris_bus_name, path, 'org.freedesktop.DBus.Properties', 'GetAll', ['org.mpris.MediaPlayer2.Player']) }}\n"
    }

    # This test is about the newline at the end of the string
    # assert to make sure it is there
    assert template["player_properties"].endswith("\n")

    res = templating.render_template(template, dict)

    assert isinstance(res, dict)
    assert isinstance(res["player_properties"], dict)
    assert res["player_properties"]["Position"] == 0

def test_nested_list_values():

    context = {
        "args": ["first-item", "second-item"]
    }
    template = {
        "res": {
            "plain_args": ["first-item", "second-item"],
            "args": "{{ args }}"
        }
    }

    templating = TemplateEngine()
    res = templating.render_template(template, dict, context)

    assert res["res"]["plain_args"] == ["first-item", "second-item"]




from datetime import datetime
from typing import Any, TypeVar

from jinja2 import BaseLoader, StrictUndefined, TemplateError
from jinja2.nativetypes import NativeEnvironment

TemplateResultType = TypeVar('TemplateResultType')

class TemplateEngine:
    def __init__(self):

        engine_globals = {}
        engine_globals['now'] = datetime.now

        self.jinja2_env = NativeEnvironment(
            loader=BaseLoader(),
            extensions=['jinja2_ansible_filters.AnsibleCoreFiltersExtension'],
            undefined=StrictUndefined,
            keep_trailing_newline=False
        )

        self.jinja2_async_env = NativeEnvironment(
            loader=BaseLoader(),
            extensions=['jinja2_ansible_filters.AnsibleCoreFiltersExtension'],
            undefined=StrictUndefined,
            enable_async=True
        )

        self.app_context: dict[str, Any] = {}

        self.jinja2_env.globals.update(engine_globals)
        self.jinja2_async_env.globals.update(engine_globals)

    def add_functions(self, custom_functions: dict[str, Any]):
        self.jinja2_env.globals.update(custom_functions)
        self.jinja2_async_env.globals.update(custom_functions)

    def update_app_context(self, context: dict[str, Any]):
        self.app_context.update(context)

    def _convert_value(self, res: Any, res_type: type[TemplateResultType]) -> TemplateResultType:

        if res is None:
            return res

        if isinstance(res, res_type):
            return res

        try:
            return res_type(res) # type: ignore

        except Exception as e:
            raise ValueError(f"Error converting rendered template result from '{type(res).__name__}' to '{res_type.__name__}'") from e

    def _render_template_nested(self, templatable: str | dict[str, Any], context: dict[str, Any] = {}) -> Any:

        if isinstance(templatable, str):
            try:
                return self.jinja2_env.from_string(templatable).render(**context)
            except TemplateError as e:
                raise TemplateError(f"Error compiling template, template={templatable}: {e}") from e

        elif isinstance(templatable, dict):
            res = {}
            for k, v in templatable.items():
                if isinstance(v, dict) or isinstance(v, str):
                    res[k] = self._render_template_nested(v, context)
                else:
                    res[k] = v
            return res

    def render_template(self, templatable: str | dict[str, Any], res_type: type[TemplateResultType], context: dict[str, Any] = {}) -> TemplateResultType:

        if isinstance(templatable, dict) and res_type is not dict:
            raise ValueError(f"res_type should dict for dictionary templates, templatable={templatable}")

        res = self._render_template_nested(templatable, context)
        res = self._convert_value(res, res_type)
        return res

    async def _async_render_template_nested(self, templatable: str | dict[str, Any], context: dict[str, Any] = {}) -> Any:

        if isinstance(templatable, str):
            try:
                return await self.jinja2_async_env.from_string(templatable).render_async(**context)
            except TemplateError as e:
                raise TemplateError(f"Error compiling template, template={templatable}: {e}") from e

        elif isinstance(templatable, dict):
            res = {}
            for k, v in templatable.items():
                if isinstance(v, dict) or isinstance(v, str):
                    res[k] = await self._async_render_template_nested(v, context)
                else:
                    res[k] = v
            return res

    async def async_render_template(self, templatable: str | dict[str, Any], res_type: type[TemplateResultType], context: dict[str, Any] = {}) -> TemplateResultType:

        if isinstance(templatable, dict) and res_type is not dict:
            raise ValueError(f"res_type should be dict for dictionary templates, templatable={templatable}")

        res = await self._async_render_template_nested(templatable, context)
        res = self._convert_value(res, res_type)
        return res





from dataclasses import dataclass

import dbus_fast.aio as dbus_aio

from dbus2mqtt.config import InterfaceConfig, SubscriptionConfig


class BusNameSubscriptions:

    def __init__(self, bus_name: str):
        self.bus_name = bus_name
        self.path_objects: dict[str, dbus_aio.proxy_object.ProxyObject] = {}

@dataclass
class SubscribedInterface:

    interface_config: InterfaceConfig
    subscription_config: SubscriptionConfig
    bus_name: str
    path: str
    interface_name: str



import fnmatch
import uuid

from dataclasses import dataclass, field
from typing import Annotated, Any, Literal

from pydantic import Field, SecretStr

from dbus2mqtt.template.templating import TemplateEngine


@dataclass
class SignalConfig:
    signal: str
    filter: str | None = None

    def matches_filter(self, template_engine: TemplateEngine, *args) -> bool:
        if self.filter:
            return template_engine.render_template(self.filter, bool, { "args": args })
        return True

@dataclass
class MethodConfig:
    method: str

@dataclass
class PropertyConfig:
    property: str

@dataclass
class InterfaceConfig:
    interface: str
    mqtt_command_topic: str | None = None
    signals: list[SignalConfig] = field(default_factory=list)
    methods: list[MethodConfig] = field(default_factory=list)
    properties: list[PropertyConfig] = field(default_factory=list)

    def render_mqtt_command_topic(self, template_engine: TemplateEngine, context: dict[str, Any]) -> Any:
        if self.mqtt_command_topic:
            return template_engine.render_template(self.mqtt_command_topic, str, context)
        return None

@dataclass
class FlowTriggerMqttConfig:
    type: Literal["mqtt"]
    topic: str
    filter: str | None = None

@dataclass
class FlowTriggerScheduleConfig:
    type: Literal["schedule"] = "schedule"
    id: str = field(default_factory=lambda: uuid.uuid4().hex)
    cron: dict[str, object] | None = None
    interval: dict[str, object] | None = None

@dataclass
class FlowTriggerDbusSignalConfig:
    interface: str
    signal: str
    type: Literal["dbus_signal"] = "dbus_signal"
    bus_name: str | None = None
    path: str | None = None
    # filter: str | None = None

@dataclass
class FlowTriggerBusNameAddedConfig:
    type: Literal["bus_name_added"] = "bus_name_added"
    # filter: str | None = None

@dataclass
class FlowTriggerBusNameRemovedConfig:
    type: Literal["bus_name_removed"] = "bus_name_removed"
    # filter: str | None = None

FlowTriggerConfig = Annotated[
    FlowTriggerMqttConfig | FlowTriggerScheduleConfig | FlowTriggerDbusSignalConfig | FlowTriggerBusNameAddedConfig | FlowTriggerBusNameRemovedConfig,
    Field(discriminator="type")
]

@dataclass
class FlowActionContextSetConfig:
    type: Literal["context_set"] = "context_set"
    context: dict[str, object] | None = None
    """Per flow execution context"""
    global_context: dict[str, object] | None = None
    """Global context, shared between multiple flow executions, over all subscriptions"""

@dataclass
class FlowActionMqttPublishConfig:
    topic: str
    payload_template: str | dict[str, Any]
    type: Literal["mqtt_publish"] = "mqtt_publish"
    payload_type: Literal["json", "yaml", "text", "binary"] = "json"

FlowActionConfig = Annotated[
    FlowActionMqttPublishConfig | FlowActionContextSetConfig,
    Field(discriminator="type")
]

@dataclass
class FlowConfig:
    triggers: list[FlowTriggerConfig]
    actions: list[FlowActionConfig]
    name: str | None = None
    id: str = field(default_factory=lambda: uuid.uuid4().hex)

@dataclass
class SubscriptionConfig:
    bus_name: str
    """bus_name pattern supporting * wildcards"""
    path: str
    """path pattern supporting * wildcards"""
    interfaces: list[InterfaceConfig] = field(default_factory=list)
    flows: list[FlowConfig] = field(default_factory=list)
    id: str = field(default_factory=lambda: uuid.uuid4().hex)

@dataclass
class DbusConfig:
    subscriptions: list[SubscriptionConfig]

    def is_bus_name_configured(self, bus_name: str) -> bool:

        for subscription in self.subscriptions:
            if fnmatch.fnmatchcase(bus_name, subscription.bus_name):
                return True
        return False

    def get_subscription_configs(self, bus_name: str, path: str) -> list[SubscriptionConfig]:
        res: list[SubscriptionConfig] = []
        for subscription in self.subscriptions:
            if fnmatch.fnmatchcase(bus_name, subscription.bus_name) and path == subscription.path:
                res.append(subscription)
        return res

@dataclass
class MqttConfig:
    host: str
    username: str
    password: SecretStr
    port: int = 1883

@dataclass
class Config:
    mqtt: MqttConfig
    dbus: DbusConfig
    flows: list[FlowConfig] = field(default_factory=list)



import pytest
import yaml

from dbus2mqtt.template.templating import TemplateEngine


def test_non_template_dict_result():

    config = yaml.safe_load("""
        payload_template:
            a_string: "Off"
            a_number: 5
    """)

    templating = TemplateEngine()
    res = templating.render_template(config["payload_template"], dict)

    assert res["a_string"] == "Off"
    assert res["a_number"] == 5

@pytest.mark.asyncio
async def test_signal_filter_bool_result():

    config = yaml.safe_load("""
        signals:
            - signal: PropertiesChanged
              filter: "{{ args[0] == 'org.mpris.MediaPlayer2.Player' }}"
    """)

    templating = TemplateEngine()
    context = {
        "args": ["org.mpris.MediaPlayer2.Player", {}, []]
    }

    res = await templating.async_render_template(config["signals"][0]["filter"], bool, context)

    assert res

@pytest.mark.asyncio
async def test_str_template_with_dict_result():
    """Test a more complex example where functions are returning dicts
    and everything must be bcomined in a nice single structure that can be converted to json"""

    config = yaml.safe_load("""
        actions:
          - type: mqtt_publish
            topic: dbus2mqtt/org.mpris.MediaPlayer2/state
            payload_template: |
                {{
                    { 'bus_name': mpris_bus_name }
                    | combine(dbus_call(mpris_bus_name, path, 'org.freedesktop.DBus.Properties', 'GetAll', ['org.mpris.MediaPlayer2.Player']))
                    | combine({ 'Volume': dbus_property_get(mpris_bus_name, path, 'org.mpris.MediaPlayer2.Player', 'Volume', 0) })
                }}
    """)

    templating = TemplateEngine()

    async def mock_dbus_property_get(bus_name, path, interface, property, default_unsupported):
        return 50

    async def mock_dbus_call(bus_name, path, interface, method, method_args):
        return {
            "CanPlay": False,
            "CanSeek": False,
            "LoopStatus": "None",
            "MaximumRate": 32.0,
            "Metadata": {
                "mpris:trackid": "/org/mpris/MediaPlayer2/firefox",
            },
            "MinimumRate": 0.032,
            "PlaybackStatus": "Stopped",
            "Position": 0,
            "Rate": 1.0,
            "Shuffle": False
        }

    context = {
        "mpris_bus_name": "org.mpris.MediaPlayer2.Player.firefox",
        "dbus_call": mock_dbus_call,
        "dbus_property_get": mock_dbus_property_get
    }

    res = await templating.async_render_template(config["actions"][0]["payload_template"], dict, context)

    assert res["bus_name"] == "org.mpris.MediaPlayer2.Player.firefox"
    assert res["Metadata"]["mpris:trackid"] == "/org/mpris/MediaPlayer2/firefox"
    assert res["Volume"] == 50



import jsonargparse


def _custom_yaml_load(stream):
    if isinstance(stream, str):
        v = stream.strip()

        # jsonargparse tries to parse yaml 1.1 boolean like values
        # Without this, str:"{'PlaybackStatus': 'Off'}" would become dict:{'PlaybackStatus': False}
        if v in ['on', 'On', 'off', 'Off', 'TRUE', 'FALSE', 'True', 'False']:
            return stream

        # Anoyingly, values starting with {{ and ending with }} are working with the default yaml_loader
        # from jsonargparse. Somehow its not when we use the custom yaml loader.
        # This fixes it
        if v.startswith("{{") or v.startswith("{%"):
            return stream

    # Delegate to default yaml loader from jsonargparse
    yaml_loader = jsonargparse.get_loader("yaml")
    return yaml_loader(stream)

def new_argument_parser() -> jsonargparse.ArgumentParser:

    # register out custom yaml loader for jsonargparse
    jsonargparse.set_loader("yaml_custom", _custom_yaml_load)

    # unless specified otherwise, load config from config.yaml
    parser = jsonargparse.ArgumentParser(default_config_files=["config.yaml"], default_env=True, env_prefix=False, parser_mode="yaml_custom")

    return parser



import dbus_fast.introspection as dbus_introspection

# taken from https://github.com/altdesktop/playerctl/blob/b19a71cb9dba635df68d271bd2b3f6a99336a223/playerctl/playerctl-daemon.c#L578
mpris_introspection_playerctl = dbus_introspection.Node.parse("""\
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
  "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
  <interface name="org.freedesktop.DBus.Introspectable">
    <method name="Introspect">
      <arg name="data" direction="out" type="s"/>
    </method>
  </interface>

  <interface name="org.freedesktop.DBus.Properties">
    <method name="Get">
      <arg direction="in" type="s"/>
      <arg direction="in" type="s"/>
      <arg direction="out" type="v"/>
    </method>
    <method name="Set">
      <arg direction="in" type="s"/>
      <arg direction="in" type="s"/>
      <arg direction="in" type="v"/>
    </method>
    <method name="GetAll">
      <arg direction="in" type="s"/>
      <arg direction="out" type="a{sv}"/>
    </method>
    <signal name="PropertiesChanged">
      <arg type="s"/>
      <arg type="a{sv}"/>
      <arg type="as"/>
    </signal>
  </interface>

  <interface name="org.mpris.MediaPlayer2">
    <property name="CanQuit" type="b" access="read"/>
    <property name="Fullscreen" type="b" access="readwrite"/>
    <property name="CanSetFullscreen" type="b" access="read"/>
    <property name="CanRaise" type="b" access="read"/>
    <property name="HasTrackList" type="b" access="read"/>
    <property name="Identity" type="s" access="read"/>
    <property name="DesktopEntry" type="s" access="read"/>
    <property name="SupportedUriSchemes" type="as" access="read"/>
    <property name="SupportedMimeTypes" type="as" access="read"/>
    <method name="Raise"/>
    <method name="Quit"/>
  </interface>

  <interface name="org.mpris.MediaPlayer2.Player">
    <property name="PlaybackStatus" type="s" access="read"/>
    <property name="LoopStatus" type="s" access="readwrite"/>
    <property name="Rate" type="d" access="readwrite"/>
    <property name="Shuffle" type="b" access="readwrite"/>
    <property name="Metadata" type="a{sv}" access="read"/>
    <property name="Volume" type="d" access="readwrite"/>
    <property name="Position" type="x" access="read"/>
    <property name="MinimumRate" type="d" access="read"/>
    <property name="MaximumRate" type="d" access="read"/>
    <property name="CanGoNext" type="b" access="read"/>
    <property name="CanGoPrevious" type="b" access="read"/>
    <property name="CanPlay" type="b" access="read"/>
    <property name="CanPause" type="b" access="read"/>
    <property name="CanSeek" type="b" access="read"/>
    <property name="CanControl" type="b" access="read"/>
    <method name="Next"/>
    <method name="Previous"/>
    <method name="Pause"/>
    <method name="PlayPause"/>
    <method name="Stop"/>
    <method name="Play"/>
    <method name="Seek">
      <arg type="x" name="Offset" direction="in"/>
    </method>
    <method name="SetPosition">
      <arg type="o" name="TrackId" direction="in"/>
      <arg type="x" name="Offset" direction="in"/>
    </method>
    <method name="OpenUri">
      <arg type="s" name="Uri" direction="in"/>
    </method>
    <signal name="Seeked">
      <arg type="x" name="Position" direction="out"/>
    </signal>
  </interface>

  <interface name="org.mpris.MediaPlayer2.TrackList">
    <property name="Tracks" type="ao" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="invalidates"/>
    </property>
    <property name="CanEditTracks" type="b" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
    </property>
    <method name="GetTracksMetadata">
      <arg direction="in" name="TrackIds" type="ao"/>
      <arg direction="out" name="Metadata" type="aa{sv}"/>
    </method>
    <method name="AddTrack">
      <arg direction="in" name="Uri" type="s"/>
      <arg direction="in" name="AfterTrack" type="o"/>
      <arg direction="in" name="SetAsCurrent" type="b"/>
    </method>
    <method name="RemoveTrack">
      <arg direction="in" name="TrackId" type="o"/>
    </method>
    <method name="GoTo">
      <arg direction="in" name="TrackId" type="o"/>
    </method>
    <signal name="TrackListReplaced">
      <arg name="Tracks" type="ao"/>
      <arg name="CurrentTrack" type="o"/>
    </signal>
    <signal name="TrackAdded">
      <arg name="Metadata" type="a{sv}"/>
      <arg name="AfterTrack" type="o"/>
    </signal>
    <signal name="TrackRemoved">
      <arg name="TrackId" type="o"/>
    </signal>
    <signal name="TrackMetadataChanged">
      <arg name="TrackId" type="o"/>
      <arg name="Metadata" type="a{sv}"/>
    </signal>
  </interface>

  <interface name="org.mpris.MediaPlayer2.Playlists">
    <method name="ActivatePlaylist">
      <arg direction="in" name="PlaylistId" type="o"/>
    </method>
    <method name="GetPlaylists">
      <arg direction="in" name="Index" type="u"/>
      <arg direction="in" name="MaxCount" type="u"/>
      <arg direction="in" name="Order" type="s"/>
      <arg direction="in" name="ReverseOrder" type="b"/>
      <arg direction="out" name="Playlists" type="a(oss)"/>
    </method>
    <property name="PlaylistCount" type="u" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
    </property>
    <property name="Orderings" type="as" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
    </property>
    <property name="ActivePlaylist" type="(b(oss))" access="read">
      <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" value="true"/>
    </property>
    <signal name="PlaylistChanged">
      <arg name="Playlist" type="(oss)"/>
    </signal>
  </interface>
</node>
""")
